from __future__ import absolute_import, unicode_literals

from mayan.apps.appearance.classes import Icon

icon_task_manager = Icon(driver_name='fontawesome', symbol='braille')
icon_queue_list = Icon(driver_name='fontawesome', symbol='braille')



from __future__ import unicode_literals

from django.utils.translation import ugettext_lazy as _

WORKFLOW_ACTION_ON_ENTRY = 1
WORKFLOW_ACTION_ON_EXIT = 2

WORKFLOW_ACTION_WHEN_CHOICES = (
    (WORKFLOW_ACTION_ON_ENTRY, _('On entry')),
    (WORKFLOW_ACTION_ON_EXIT, _('On exit')),
)



from __future__ import unicode_literals

from django import forms
from django.utils.encoding import force_text
from django.utils.translation import ugettext_lazy as _

from mayan.apps.common.classes import ModelField, ModelProperty
from mayan.apps.documents.models import Document

from .models import SmartLink, SmartLinkCondition


class SmartLinkForm(forms.ModelForm):
    def __init__(self, *args, **kwargs):
        super(SmartLinkForm, self).__init__(*args, **kwargs)
        self.fields['dynamic_label'].help_text = ' '.join(
            [
                force_text(self.fields['dynamic_label'].help_text),
                ModelProperty.get_help_text_for(
                    model=Document, show_name=True
                ).replace('\n', '<br>')
            ]
        )

    class Meta:
        fields = ('label', 'dynamic_label', 'enabled')
        model = SmartLink


class SmartLinkConditionForm(forms.ModelForm):
    def __init__(self, *args, **kwargs):
        super(SmartLinkConditionForm, self).__init__(*args, **kwargs)
        self.fields['foreign_document_data'] = forms.ChoiceField(
            choices=ModelField.get_choices_for(
                model=Document,
            ), label=_('Foreign document field')
        )
        self.fields['expression'].help_text = ' '.join(
            [
                force_text(self.fields['expression'].help_text),
                ModelProperty.get_help_text_for(
                    model=Document, show_name=True
                ).replace('\n', '<br>')
            ]
        )

    class Meta:
        model = SmartLinkCondition
        exclude = ('smart_link',)



from __future__ import unicode_literals

from django.template import Library

from ..classes import Link, Menu, SourceColumn
from ..forms import MultiItemForm

register = Library()


@register.simple_tag()
def get_link(name):
    return Link.get(name)


@register.simple_tag()
def get_menu(name):
    return Menu.get(name)


@register.simple_tag(takes_context=True)
def get_menu_links(context, name, source=None, sort_results=None):
    return Menu.get(name).resolve(context=context, source=source, sort_results=sort_results)


@register.simple_tag(takes_context=True)
def get_menus_links(context, names, source=None, sort_results=None):
    result = []

    for name in names.split(','):
        for links in Menu.get(name=name).resolve(context=context, sort_results=sort_results):
            if links:
                result.append(links)

    return result


@register.simple_tag(takes_context=True)
def get_multi_item_links_form(context, object_list):
    actions = []
    for link_set in Menu.get('multi item').resolve(context=context, source=object_list[0], sort_results=True):
        for link in link_set['links']:
            if hasattr(link, 'url'):
                actions.append((link.url, link.text))

    form = MultiItemForm(actions=actions)
    context.update({'multi_item_form': form, 'multi_item_actions': actions})
    return ''


@register.simple_tag(takes_context=True)
def navigation_get_sort_field_querystring(context, column):
    return column.get_sort_field_querystring(context=context)


@register.simple_tag(takes_context=True)
def navigation_get_source_columns(context, source, exclude_identifier=False, only_identifier=False):
    try:
        # Is it a query set?
        source = source.model
    except AttributeError:
        # Is not a query set
        try:
            # Is iterable?
            source = source[0]
        except TypeError:
            # It is not an iterable
            pass
        except IndexError:
            # It a list and it's empty
            pass
        except KeyError:
            # It a list and it's empty
            pass

    return SourceColumn.get_for_source(
        context=context, source=source, exclude_identifier=exclude_identifier,
        only_identifier=only_identifier
    )


@register.simple_tag(takes_context=True)
def navigation_resolve_menu(context, name, source=None, sort_results=None):
    result = []

    menu = Menu.get(name)
    link_groups = menu.resolve(
        context=context, source=source, sort_results=sort_results
    )

    if link_groups:
        result.append({'link_groups': link_groups, 'menu': menu})

    return result


@register.simple_tag(takes_context=True)
def navigation_resolve_menus(context, names, source=None, sort_results=None):
    result = []

    for name in names.split(','):
        menu = Menu.get(name=name)
        link_groups = menu.resolve(context=context, sort_results=sort_results)

        if link_groups:
            result.append({'link_groups': link_groups, 'menu': menu})

    return result


@register.simple_tag()
def navigation_source_column_get_absolute_url(source_column, obj):
    return source_column.get_absolute_url(obj=obj)


@register.simple_tag(takes_context=True)
def resolve_link(context, link):
    # This can be used to resolve links or menus too
    return link.resolve(context=context)


@register.simple_tag(takes_context=True)
def navigation_source_column_resolve(context, column):
    if column:
        result = column.resolve(context=context)
        return result
    else:
        return ''



from __future__ import unicode_literals

from django.conf.urls import url

from .views import (
    QueueListView, QueueActiveTaskListView, QueueScheduledTaskListView,
    QueueReservedTaskListView
)


urlpatterns = [
    url(
        regex=r'^queues/$', view=QueueListView.as_view(),
        name='queue_list'
    ),
    url(
        regex=r'^queues/(?P<queue_name>[-\w]+)/tasks/active/$',
        view=QueueActiveTaskListView.as_view(), name='queue_active_task_list'
    ),
    url(
        regex=r'^queues/(?P<queue_name>[-\w]+)/tasks/scheduled/$',
        view=QueueScheduledTaskListView.as_view(),
        name='queue_scheduled_task_list'
    ),
    url(
        regex=r'^queues/(?P<queue_name>[-\w]+)/tasks/reserved/$',
        view=QueueReservedTaskListView.as_view(),
        name='queue_reserved_task_list'
    ),
]



from __future__ import unicode_literals

from .literals import TEST_EMAIL_BASE64_FILENAME


class MockIMAPServer(object):
    def login(self, user, password):
        return ('OK', ['{} authenticated (Success)'.format(user)])

    def select(self, mailbox='INBOX', readonly=False):
        return ('OK', ['1'])

    def search(self, charset, *criteria):
        return ('OK', ['1'])

    def fetch(self, message_set, message_parts):
        return (
            'OK', [
                (
                    '1 (RFC822 {4800}',
                    TEST_EMAIL_BASE64_FILENAME
                ), ' FLAGS (\\Seen))'
            ]
        )

    def store(self, message_set, command, flags):
        return ('OK', ['1 (FLAGS (\\Seen \\Deleted))'])

    def expunge(self):
        return ('OK', ['1'])

    def close(self):
        return ('OK', ['Returned to authenticated state. (Success)'])

    def logout(self):
        return ('BYE', ['LOGOUT Requested'])


class MockMailbox(object):
    def dele(self, which):
        return

    def getwelcome(self):
        return

    def list(self, which=None):
        return (None, ['1 test'])

    def user(self, user):
        return

    def pass_(self, pswd):
        return

    def quit(self):
        return

    def retr(self, which=None):
        return (
            1, [TEST_EMAIL_BASE64_FILENAME]
        )


class MockStagingFolder(object):
    """Mock of a StagingFolder model"""



from __future__ import unicode_literals

from mayan.apps.common.tests import BaseTestCase
from mayan.apps.document_indexing.models import Index, IndexInstanceNode
from mayan.apps.document_indexing.tests.literals import TEST_INDEX_LABEL
from mayan.apps.documents.tests import DocumentTestMixin

from .literals import (
    TEST_FILE_METADATA_INDEX_NODE_TEMPLATE, TEST_FILE_METADATA_VALUE
)


class IndexingTestCase(DocumentTestMixin, BaseTestCase):
    auto_upload_document = False

    def test_indexing(self):
        index = Index.objects.create(label=TEST_INDEX_LABEL)

        index.document_types.add(self.test_document_type)

        root = index.template_root
        index.node_templates.create(
            parent=root, expression=TEST_FILE_METADATA_INDEX_NODE_TEMPLATE,
            link_documents=True
        )
        self.upload_document()
        self.test_document.submit_for_file_metadata_processing()
        index.rebuild()
        self.assertTrue(
            self.test_document in IndexInstanceNode.objects.get(
                value=TEST_FILE_METADATA_VALUE
            ).documents.all()
        )



from __future__ import unicode_literals

from actstream.models import Action

from mayan.apps.common.tests import GenericViewTestCase
from mayan.apps.documents.tests import DocumentTestMixin

from ..permissions import (
    permission_smart_link_create, permission_smart_link_edit,
)

from ..events import event_smart_link_created, event_smart_link_edited

from .mixins import SmartLinkTestMixin, SmartLinkViewTestMixin


class SmartLinkTemplateEventsTestCase(DocumentTestMixin, SmartLinkTestMixin, SmartLinkViewTestMixin, GenericViewTestCase):
    auto_upload_document = False

    def test_smart_link_create_event(self):
        self.grant_permission(
            permission=permission_smart_link_create
        )
        Action.objects.all().delete()

        response = self._request_test_smart_link_create_view()
        self.assertEqual(response.status_code, 302)

        action = Action.objects.last()

        self.assertEqual(action.actor, self._test_case_user)
        self.assertEqual(action.target, self.test_smart_link)
        self.assertEqual(action.verb, event_smart_link_created.id)

    def test_smart_link_edit_event(self):
        self._create_test_smart_link()

        self.grant_access(
            obj=self.test_smart_link, permission=permission_smart_link_edit
        )
        Action.objects.all().delete()

        response = self._request_test_smart_link_edit_view()
        self.assertEqual(response.status_code, 302)

        action = Action.objects.last()

        self.assertEqual(action.actor, self._test_case_user)
        self.assertEqual(action.target, self.test_smart_link)
        self.assertEqual(action.verb, event_smart_link_edited.id)



from __future__ import absolute_import, unicode_literals

from django.apps import apps
from django.db.models.signals import m2m_changed, pre_delete
from django.utils.translation import ugettext_lazy as _

from mayan.apps.acls.classes import ModelPermission
from mayan.apps.acls.links import link_acl_list
from mayan.apps.acls.permissions import permission_acl_edit, permission_acl_view
from mayan.apps.common.apps import MayanAppConfig
from mayan.apps.common.classes import ModelField
from mayan.apps.common.menus import (
    menu_facet, menu_list_facet, menu_main, menu_multi_item, menu_object,
    menu_secondary
)
from mayan.apps.documents.search import document_page_search, document_search
from mayan.apps.events.classes import ModelEventType
from mayan.apps.events.links import (
    link_events_for_object, link_object_event_types_user_subcriptions_list,
)
from mayan.apps.events.permissions import permission_events_view
from mayan.apps.navigation.classes import SourceColumn

from .dependencies import *  # NOQA
from .events import (
    event_tag_attach, event_tag_edited, event_tag_remove
)
from .handlers import handler_index_document, handler_tag_pre_delete
from .html_widgets import widget_document_tags
from .links import (
    link_document_tag_list, link_document_multiple_attach_multiple_tag,
    link_document_multiple_tag_multiple_remove,
    link_document_tag_multiple_remove, link_document_tag_multiple_attach, link_tag_create,
    link_tag_delete, link_tag_edit, link_tag_list,
    link_tag_multiple_delete, link_tag_document_list
)
from .menus import menu_tags
from .methods import method_document_get_tags
from .permissions import (
    permission_tag_attach, permission_tag_delete, permission_tag_edit,
    permission_tag_remove, permission_tag_view
)
from .search import tag_search  # NOQA


class TagsApp(MayanAppConfig):
    app_namespace = 'tags'
    app_url = 'tags'
    has_rest_api = True
    has_tests = True
    name = 'mayan.apps.tags'
    verbose_name = _('Tags')

    def ready(self):
        super(TagsApp, self).ready()
        from actstream import registry

        from .wizard_steps import WizardStepTags  # NOQA

        Document = apps.get_model(
            app_label='documents', model_name='Document'
        )

        DocumentPageResult = apps.get_model(
            app_label='documents', model_name='DocumentPageResult'
        )

        DocumentTag = self.get_model(model_name='DocumentTag')
        Tag = self.get_model(model_name='Tag')

        Document.add_to_class(name='get_tags', value=method_document_get_tags)

        ModelEventType.register(
            model=Tag, event_types=(
                event_tag_attach, event_tag_edited, event_tag_remove
            )
        )

        ModelField(model=Document, name='tags__label')
        ModelField(model=Document, name='tags__color')

        ModelPermission.register(
            model=Document, permissions=(
                permission_tag_attach, permission_tag_remove,
                permission_tag_view
            )
        )

        ModelPermission.register(
            model=Tag, permissions=(
                permission_acl_edit, permission_acl_view,
                permission_events_view, permission_tag_attach,
                permission_tag_delete, permission_tag_edit,
                permission_tag_remove, permission_tag_view,
            )
        )

        SourceColumn(
            attribute='label', is_identifier=True, is_sortable=True,
            source=DocumentTag
        )
        SourceColumn(
            attribute='get_preview_widget', source=DocumentTag
        )

        SourceColumn(
            source=Document, label=_('Tags'),
            func=lambda context: widget_document_tags(
                document=context['object'], user=context['request'].user
            )
        )

        SourceColumn(
            source=DocumentPageResult, label=_('Tags'),
            func=lambda context: widget_document_tags(
                document=context['object'].document,
                user=context['request'].user
            )
        )

        SourceColumn(
            attribute='label', is_identifier=True, is_sortable=True,
            source=Tag
        )
        SourceColumn(
            attribute='get_preview_widget', source=Tag
        )
        SourceColumn(
            source=Tag, label=_('Documents'),
            func=lambda context: context['object'].get_document_count(
                user=context['request'].user
            )
        )

        document_page_search.add_model_field(
            field='document_version__document__tags__label', label=_('Tags')
        )
        document_search.add_model_field(field='tags__label', label=_('Tags'))

        menu_facet.bind_links(
            links=(link_document_tag_list,), sources=(Document,)
        )

        menu_list_facet.bind_links(
            links=(
                link_acl_list, link_events_for_object,
                link_object_event_types_user_subcriptions_list,
                link_tag_document_list,
            ), sources=(Tag,)
        )

        menu_tags.bind_links(
            links=(
                link_tag_list, link_tag_create
            )
        )

        menu_main.bind_links(links=(menu_tags,), position=98)

        menu_multi_item.bind_links(
            links=(
                link_document_multiple_attach_multiple_tag,
                link_document_multiple_tag_multiple_remove
            ),
            sources=(Document,)
        )
        menu_multi_item.bind_links(
            links=(link_tag_multiple_delete,), sources=(Tag,)
        )
        menu_object.bind_links(
            links=(
                link_tag_edit, link_tag_delete
            ),
            sources=(Tag,)
        )
        menu_secondary.bind_links(
            links=(link_document_tag_multiple_attach, link_document_tag_multiple_remove),
            sources=(
                'tags:tag_attach', 'tags:document_tag_list',
                'tags:single_document_multiple_tag_remove'
            )
        )
        registry.register(Tag)

        # Index update

        m2m_changed.connect(
            dispatch_uid='tags_handler_index_document',
            receiver=handler_index_document,
            sender=Tag.documents.through
        )

        pre_delete.connect(
            dispatch_uid='tags_handler_tag_pre_delete',
            receiver=handler_tag_pre_delete,
            sender=Tag
        )



from __future__ import unicode_literals

import errno
import os

import yaml


def read_configuration_file(path):
    try:
        with open(path) as file_object:
            file_object.seek(0, os.SEEK_END)
            if file_object.tell():
                file_object.seek(0)
                try:
                    return yaml.safe_load(file_object)
                except yaml.YAMLError as exception:
                    exit(
                        'Error loading configuration file: {}; {}'.format(
                            path, exception
                        )
                    )
    except IOError as exception:
        if exception.errno == errno.ENOENT:
            pass
        else:
            raise


def yaml_loads(data, error_message=None):
    if not error_message:
        error_message = 'Error loading: {}; {}'

    try:
        return yaml.safe_load(data)
    except yaml.YAMLError as exception:
        exit(
            error_message.format(data, exception)
        )



from __future__ import unicode_literals

import yaml

try:
    from yaml import CSafeLoader as SafeLoader
except ImportError:
    from yaml import SafeLoader

from django.utils.module_loading import import_string

from .settings import (
    setting_documentimagecache_storage,
    setting_documentimagecache_storage_arguments,
    setting_storage_backend, setting_storage_backend_arguments
)

storage_documentversion = import_string(
    dotted_path=setting_storage_backend.value
)(
    **yaml.load(
        stream=setting_storage_backend_arguments.value or '{}',
        Loader=SafeLoader
    )
)

storage_documentimagecache = import_string(
    dotted_path=setting_documentimagecache_storage.value
)(
    **yaml.load(
        stream=setting_documentimagecache_storage_arguments.value or '{}',
        Loader=SafeLoader
    )
)



from __future__ import absolute_import, unicode_literals

from mayan.apps.appearance.classes import Icon
from mayan.apps.documents.icons import icon_document, icon_document_type


icon_workflow = Icon(driver_name='fontawesome', symbol='sitemap')

icon_document_type_workflow_list = icon_workflow

icon_document_workflow_instance_list = Icon(
    driver_name='fontawesome', symbol='sitemap'
)
icon_setup_workflow_list = Icon(driver_name='fontawesome', symbol='sitemap')
icon_tool_launch_all_workflows = Icon(
    driver_name='fontawesome', symbol='sitemap'
)
icon_workflow_create = Icon(
    driver_name='fontawesome-dual', primary_symbol='sitemap',
    secondary_symbol='plus'
)
icon_workflow_delete = Icon(driver_name='fontawesome', symbol='times')
icon_workflow_document_type_list = icon_document_type
icon_workflow_edit = Icon(driver_name='fontawesome', symbol='pencil-alt')
icon_workflow_list = Icon(driver_name='fontawesome', symbol='sitemap')
icon_workflow_preview = Icon(driver_name='fontawesome', symbol='eye')


icon_workflow_instance_detail = Icon(driver_name='fontawesome', symbol='sitemap')
icon_workflow_instance_transition = Icon(
    driver_name='fontawesome', symbol='arrows-alt-h'
)

icon_workflow_runtime_proxy_document_list = icon_document
icon_workflow_runtime_proxy_list = Icon(
    driver_name='fontawesome', symbol='sitemap'
)
icon_workflow_runtime_proxy_state_document_list = icon_document
icon_workflow_runtime_proxy_state_list = Icon(
    driver_name='fontawesome', symbol='circle'
)

icon_workflow_state_action_delete = Icon(
    driver_name='fontawesome', symbol='times'
)
icon_workflow_state_action_edit = Icon(
    driver_name='fontawesome', symbol='pencil-alt'
)
icon_workflow_state_action_list = Icon(
    driver_name='fontawesome', symbol='code'
)
icon_workflow_state = Icon(driver_name='fontawesome', symbol='circle')
icon_workflow_state_create = Icon(
    driver_name='fontawesome-dual', primary_symbol='circle',
    secondary_symbol='plus'
)
icon_workflow_state_delete = Icon(driver_name='fontawesome', symbol='times')
icon_workflow_state_edit = Icon(driver_name='fontawesome', symbol='pencil-alt')

icon_workflow_state_action = Icon(driver_name='fontawesome', symbol='code')
icon_workflow_state_action_delete = Icon(driver_name='fontawesome', symbol='times')
icon_workflow_state_action_edit = Icon(driver_name='fontawesome', symbol='pencil-alt')
icon_workflow_state_action_selection = Icon(
    driver_name='fontawesome-dual', primary_symbol='code',
    secondary_symbol='plus'
)
icon_workflow_state_action_list = Icon(driver_name='fontawesome', symbol='code')
icon_workflow_transition = Icon(
    driver_name='fontawesome', symbol='arrows-alt-h'
)
icon_workflow_transition_create = Icon(
    driver_name='fontawesome-dual', primary_symbol='arrows-alt-h',
    secondary_symbol='plus'
)
icon_workflow_transition_delete = Icon(driver_name='fontawesome', symbol='times')
icon_workflow_transition_edit = Icon(
    driver_name='fontawesome', symbol='pencil-alt'
)
icon_workflow_transition_triggers = Icon(
    driver_name='fontawesome', symbol='bolt'
)



from __future__ import absolute_import, unicode_literals

import hashlib

from fuse import FuseOSError

from mayan.apps.common.tests import BaseTestCase
from mayan.apps.documents.models import Document
from mayan.apps.documents.tests import DocumentTestMixin
from mayan.apps.document_indexing.tests import IndexTestMixin

from ..filesystems import IndexFilesystem

from .literals import (
    TEST_NODE_EXPRESSION, TEST_NODE_EXPRESSION_MULTILINE,
    TEST_NODE_EXPRESSION_MULTILINE_EXPECTED, TEST_NODE_EXPRESSION_MULTILINE_2,
    TEST_NODE_EXPRESSION_MULTILINE_2_EXPECTED
)


class IndexFilesystemTestCase(IndexTestMixin, DocumentTestMixin, BaseTestCase):
    auto_upload_document = False

    def test_document_access(self):
        self._create_test_index()

        self.test_index.node_templates.create(
            parent=self.test_index.template_root, expression=TEST_NODE_EXPRESSION,
            link_documents=True
        )

        self.upload_document()
        index_filesystem = IndexFilesystem(index_slug=self.test_index.slug)

        self.assertEqual(
            index_filesystem.access(
                '/{}/{}'.format(TEST_NODE_EXPRESSION, self.test_document.label)
            ), None
        )

    def test_document_access_failure(self):
        self._create_test_index()

        self.test_index.node_templates.create(
            parent=self.test_index.template_root, expression=TEST_NODE_EXPRESSION,
            link_documents=True
        )

        self.upload_document()
        index_filesystem = IndexFilesystem(index_slug=self.test_index.slug)

        with self.assertRaises(FuseOSError):
            index_filesystem.access(
                '/{}/{}_non_valid'.format(TEST_NODE_EXPRESSION, self.test_document.label)
            )

    def test_document_open(self):
        self._create_test_index()

        self.test_index.node_templates.create(
            parent=self.test_index.template_root, expression=TEST_NODE_EXPRESSION,
            link_documents=True
        )

        self.upload_document()
        index_filesystem = IndexFilesystem(index_slug=self.test_index.slug)

        file_handle = index_filesystem.open(
            '/{}/{}'.format(TEST_NODE_EXPRESSION, self.test_document.label),
            'rb'
        )

        self.assertEqual(
            hashlib.sha256(
                index_filesystem.read(
                    fh=file_handle, offset=0, path=None,
                    size=self.test_document.size
                )
            ).hexdigest(),
            self.test_document.checksum
        )

    def test_multiline_indexes(self):
        self._create_test_index()

        self.test_index.node_templates.create(
            parent=self.test_index.template_root,
            expression=TEST_NODE_EXPRESSION_MULTILINE,
            link_documents=True
        )

        self.upload_document()
        index_filesystem = IndexFilesystem(index_slug=self.test_index.slug)

        self.assertEqual(
            list(index_filesystem.readdir('/', ''))[2:],
            [TEST_NODE_EXPRESSION_MULTILINE_EXPECTED]
        )

    def test_multiline_indexes_first_and_last(self):
        self._create_test_index()

        self.test_index.node_templates.create(
            parent=self.test_index.template_root,
            expression=TEST_NODE_EXPRESSION_MULTILINE_2,
            link_documents=True
        )

        self.upload_document()
        index_filesystem = IndexFilesystem(index_slug=self.test_index.slug)

        self.assertEqual(
            list(index_filesystem.readdir('/', ''))[2:],
            [TEST_NODE_EXPRESSION_MULTILINE_2_EXPECTED]
        )

    def test_duplicated_indexes(self):
        self._create_test_index()

        self.test_index.node_templates.create(
            parent=self.test_index.template_root, expression=TEST_NODE_EXPRESSION,
            link_documents=True
        )
        self.test_index.node_templates.create(
            parent=self.test_index.template_root, expression=TEST_NODE_EXPRESSION,
            link_documents=True
        )

        self.upload_document()
        index_filesystem = IndexFilesystem(index_slug=self.test_index.slug)

        self.assertEqual(
            list(index_filesystem.readdir('/', ''))[2:], []
        )

    def test_ignore_stub_documents(self):
        self._create_test_index()

        self.test_index.node_templates.create(
            parent=self.test_index.template_root, expression=TEST_NODE_EXPRESSION,
            link_documents=True
        )

        self.test_document = Document.objects.create(
            document_type=self.test_document_type, label='document_stub'
        )
        index_filesystem = IndexFilesystem(index_slug=self.test_index.slug)
        self.test_index.rebuild()

        self.assertEqual(
            list(
                index_filesystem.readdir('/', ''))[2:], []
        )



from __future__ import absolute_import, unicode_literals

from django.contrib import messages
from django.http import HttpResponseRedirect
from django.shortcuts import get_object_or_404
from django.urls import reverse
from django.utils.translation import ugettext_lazy as _

from mayan.apps.acls.models import AccessControlList
from mayan.apps.common.generics import (
    ConfirmView, SingleObjectCreateView, SingleObjectDetailView
)
from mayan.apps.common.utils import encapsulate
from mayan.apps.documents.models import Document
from mayan.apps.documents.views import DocumentListView

from .exceptions import DocumentAlreadyCheckedOut, DocumentNotCheckedOut
from .forms import DocumentCheckoutForm, DocumentCheckoutDefailForm
from .icons import icon_check_out_info
from .models import DocumentCheckout
from .permissions import (
    permission_document_check_in, permission_document_check_in_override,
    permission_document_check_out, permission_document_check_out_detail_view
)


class DocumentCheckInView(ConfirmView):
    def get_extra_context(self):
        document = self.get_object()

        context = {
            'object': document,
        }

        if document.get_check_out_info().user != self.request.user:
            context['title'] = _(
                'You didn\'t originally checked out this document. '
                'Forcefully check in the document: %s?'
            ) % document
        else:
            context['title'] = _('Check in the document: %s?') % document

        return context

    def get_object(self):
        return get_object_or_404(klass=Document, pk=self.kwargs['pk'])

    def get_post_action_redirect(self):
        return reverse(
            viewname='checkouts:check_out_info', kwargs={
                'pk': self.get_object().pk
            }
        )

    def view_action(self):
        document = self.get_object()

        if document.get_check_out_info().user == self.request.user:
            AccessControlList.objects.check_access(
                obj=document, permissions=(permission_document_check_in,),
                user=self.request.user
            )
        else:
            AccessControlList.objects.check_access(
                obj=document,
                permissions=(permission_document_check_in_override,),
                user=self.request.user
            )

        try:
            document.check_in(user=self.request.user)
        except DocumentNotCheckedOut:
            messages.error(
                message=_('Document has not been checked out.'),
                request=self.request
            )
        else:
            messages.success(
                message=_(
                    'Document "%s" checked in successfully.'
                ) % document, request=self.request
            )


class DocumentCheckOutDetailView(SingleObjectDetailView):
    form_class = DocumentCheckoutDefailForm
    model = Document
    object_permission = permission_document_check_out_detail_view

    def get_extra_context(self):
        return {
            'object': self.object,
            'title': _(
                'Check out details for document: %s'
            ) % self.object
        }


class DocumentCheckOutView(SingleObjectCreateView):
    form_class = DocumentCheckoutForm

    def dispatch(self, request, *args, **kwargs):
        self.document = get_object_or_404(klass=Document, pk=self.kwargs['pk'])

        AccessControlList.objects.check_access(
            obj=self.document, permissions=(permission_document_check_out,),
            user=request.user
        )

        return super(
            DocumentCheckOutView, self
        ).dispatch(request, *args, **kwargs)

    def form_valid(self, form):
        try:
            instance = form.save(commit=False)
            instance.user = self.request.user
            instance.document = self.document
            instance.save()
        except DocumentAlreadyCheckedOut:
            messages.error(
                message=_('Document already checked out.'),
                request=self.request
            )
        else:
            messages.success(
                message=_(
                    'Document "%s" checked out successfully.'
                ) % self.document, request=self.request
            )

        return HttpResponseRedirect(redirect_to=self.get_success_url())

    def get_extra_context(self):
        return {
            'object': self.document,
            'title': _('Check out document: %s') % self.document
        }

    def get_post_action_redirect(self):
        return reverse(
            viewname='checkouts:check_out_info', kwargs={
                'pk': self.document.pk
            }
        )


class DocumentCheckOutListView(DocumentListView):
    def get_document_queryset(self):
        return AccessControlList.objects.restrict_queryset(
            permission=permission_document_check_out_detail_view,
            queryset=DocumentCheckout.objects.checked_out_documents(),
            user=self.request.user
        )

    def get_extra_context(self):
        context = super(DocumentCheckOutListView, self).get_extra_context()
        context.update(
            {
                'extra_columns': (
                    {
                        'name': _('User'),
                        'attribute': encapsulate(
                            lambda document: document.get_check_out_info().user.get_full_name() or document.get_check_out_info().user
                        )
                    },
                    {
                        'name': _('Checkout time and date'),
                        'attribute': encapsulate(
                            lambda document: document.get_check_out_info().checkout_datetime
                        )
                    },
                    {
                        'name': _('Checkout expiration'),
                        'attribute': encapsulate(
                            lambda document: document.get_check_out_info().expiration_datetime
                        )
                    },
                ),
                'no_results_icon': icon_check_out_info,
                'no_results_text': _(
                    'Checking out a document blocks certain document '
                    'operations for a predetermined amount of '
                    'time.'
                ),
                'no_results_title': _('No documents have been checked out'),
                'title': _('Documents checked out'),
            }
        )
        return context



from __future__ import absolute_import, unicode_literals

from . import *  # NOQA

ALLOWED_HOSTS = ['*']

DEBUG = True

CELERY_ALWAYS_EAGER = True
CELERY_EAGER_PROPAGATES_EXCEPTIONS = CELERY_ALWAYS_EAGER

EMAIL_BACKEND = 'django.core.mail.backends.console.EmailBackend'


if 'rosetta' not in INSTALLED_APPS:
    try:
        import rosetta
    except ImportError:
        pass
    else:
        INSTALLED_APPS += (
            'rosetta',
        )

if 'django_extensions' not in INSTALLED_APPS:
    try:
        import django_extensions
    except ImportError:
        pass
    else:
        INSTALLED_APPS += (
            'django_extensions',
        )

ROOT_URLCONF = 'mayan.urls.development'

TEMPLATES[0]['OPTIONS']['loaders'] = (
    'django.template.loaders.filesystem.Loader',
    'django.template.loaders.app_directories.Loader',
)

WSGI_AUTO_RELOAD = True



from __future__ import unicode_literals

from django.contrib.contenttypes.models import ContentType

from mayan.apps.documents.tests import GenericDocumentViewTestCase

from ..models import Transformation
from ..permissions import (
    permission_transformation_create, permission_transformation_delete,
    permission_transformation_view
)

from .literals import TEST_TRANSFORMATION_NAME, TEST_TRANSFORMATION_ARGUMENT


class TransformationViewsTestCase(GenericDocumentViewTestCase):
    def _transformation_create_view(self):
        return self.post(
            viewname='converter:transformation_create', kwargs={
                'app_label': 'documents', 'model': 'document',
                'object_id': self.test_document.pk
            }, data={
                'name': TEST_TRANSFORMATION_NAME,
                'arguments': TEST_TRANSFORMATION_ARGUMENT
            }
        )

    def test_transformation_create_view_no_permissions(self):
        transformation_count = Transformation.objects.count()

        response = self._transformation_create_view()
        self.assertEqual(response.status_code, 403)

        self.assertEqual(Transformation.objects.count(), transformation_count)

    def test_transformation_create_view_with_permissions(self):
        self.grant_permission(permission=permission_transformation_create)

        transformation_count = Transformation.objects.count()

        response = self._transformation_create_view()
        self.assertEqual(response.status_code, 302)

        self.assertEqual(
            Transformation.objects.count(), transformation_count + 1
        )

    def _request_transformation_delete_view(self):
        return self.post(
            viewname='converter:transformation_delete', kwargs={
                'pk': self.test_transformation.pk
            }
        )

    def _create_test_transformation(self):
        content_type = ContentType.objects.get_for_model(model=self.test_document)

        self.test_transformation = Transformation.objects.create(
            content_type=content_type, object_id=self.test_document.pk,
            name=TEST_TRANSFORMATION_NAME,
            arguments=TEST_TRANSFORMATION_ARGUMENT
        )

    def test_transformation_delete_view_no_permissions(self):
        self._create_test_transformation()

        transformation_count = Transformation.objects.count()

        response = self._request_transformation_delete_view()
        self.assertEqual(response.status_code, 403)

        self.assertEqual(
            Transformation.objects.count(), transformation_count
        )

    def test_transformation_delete_view_with_permissions(self):
        self._create_test_transformation()

        self.grant_permission(permission=permission_transformation_delete)

        transformation_count = Transformation.objects.count()

        response = self._request_transformation_delete_view()
        self.assertEqual(response.status_code, 302)

        self.assertEqual(
            Transformation.objects.count(), transformation_count - 1
        )

    def _transformation_list_view(self):
        return self.get(
            viewname='converter:transformation_list', kwargs={
                'app_label': 'documents', 'model': 'document',
                'object_id': self.test_document.pk
            }
        )

    def test_transformation_list_view_no_permissions(self):
        response = self._transformation_list_view()
        self.assertEqual(response.status_code, 403)

    def test_transformation_list_view_with_permissions(self):
        self.grant_access(
            obj=self.test_document, permission=permission_transformation_view
        )

        response = self._transformation_list_view()
        self.assertContains(
            response=response, text=self.test_document.label, status_code=200
        )



from __future__ import absolute_import, unicode_literals

from ..classes import Namespace

from .literals import TEST_NAMESPACE_LABEL, TEST_NAMESPACE_NAME


class SmartSettingsTestCaseMixin(object):
    def setUp(self):
        super(SmartSettingsTestCaseMixin, self).setUp()
        Namespace.invalidate_cache_all()


class SmartSettingTestMixin(object):
    def _create_test_settings_namespace(self):
        try:
            self.test_settings_namespace = Namespace.get(
                name=TEST_NAMESPACE_NAME
            )
        except KeyError:
            self.test_settings_namespace = Namespace(
                label=TEST_NAMESPACE_LABEL, name=TEST_NAMESPACE_NAME
            )



from __future__ import unicode_literals

import itertools
import logging

from django.apps import apps
from django.core.exceptions import PermissionDenied
from django.utils.encoding import force_text, python_2_unicode_compatible
from django.utils.translation import ugettext_lazy as _

from .exceptions import InvalidNamespace

logger = logging.getLogger(__name__)


@python_2_unicode_compatible
class PermissionNamespace(object):
    _registry = {}

    @classmethod
    def all(cls):
        return cls._registry.values()

    @classmethod
    def get(cls, name):
        try:
            return cls._registry[name]
        except KeyError:
            raise InvalidNamespace(
                'Invalid namespace name. This is probably an obsolete '
                'permission namespace, execute the management command '
                '"purgepermissions" and try again.'
            )

    def __init__(self, name, label):
        self.name = name
        self.label = label
        self.permissions = []
        self.__class__._registry[name] = self

    def __str__(self):
        return force_text(self.label)

    def add_permission(self, name, label):
        permission = Permission(namespace=self, name=name, label=label)
        self.permissions.append(permission)
        return permission


@python_2_unicode_compatible
class Permission(object):
    _permissions = {}
    _stored_permissions_cache = {}

    @classmethod
    def all(cls, as_choices=False):
        if as_choices:
            results = []

            for namespace, permissions in itertools.groupby(cls.all(), lambda entry: entry.namespace):
                permission_options = [
                    (force_text(permission.pk), permission) for permission in permissions
                ]
                results.append(
                    (namespace, permission_options)
                )

            return results
        else:
            # Return sorted permisions by namespace.name
            return sorted(
                cls._permissions.values(), key=lambda x: x.namespace.name
            )

    @classmethod
    def check_user_permissions(cls, permissions, user):
        for permission in permissions:
            if permission.stored_permission.user_has_this(user=user):
                return True

        logger.debug(
            'User "%s" does not have permissions "%s"', user, permissions
        )
        raise PermissionDenied(_('Insufficient permissions.'))

    @classmethod
    def get(cls, pk, proxy_only=False):
        if proxy_only:
            return cls._permissions[pk]
        else:
            return cls._permissions[pk].stored_permission

    @classmethod
    def invalidate_cache(cls):
        cls._stored_permissions_cache = {}

    @classmethod
    def refresh(cls):
        for permission in cls.all():
            permission.stored_permission

    def __init__(self, namespace, name, label):
        self.namespace = namespace
        self.name = name
        self.label = label
        self.pk = self.get_pk()
        self.__class__._permissions[self.pk] = self

    def __repr__(self):
        return self.pk

    def __str__(self):
        return force_text(self.label)

    def get_pk(self):
        return '%s.%s' % (self.namespace.name, self.name)

    @property
    def stored_permission(self):
        try:
            return self.__class__._stored_permissions_cache[self.pk]
        except KeyError:
            StoredPermission = apps.get_model(
                app_label='permissions', model_name='StoredPermission'
            )

            stored_permission, created = StoredPermission.objects.get_or_create(
                namespace=self.namespace.name,
                name=self.name,
            )

            self.__class__._stored_permissions_cache[self.pk] = stored_permission
            return stored_permission



from __future__ import unicode_literals

from collections import OrderedDict

from django import forms
from django.utils.safestring import mark_safe


class DisableableSelectWidget(forms.widgets.SelectMultiple):
    def create_option(self, *args, **kwargs):
        result = super(DisableableSelectWidget, self).create_option(*args, **kwargs)

        # Get a keyword argument named value or the second positional argument
        # Current interface as of Django 1.11
        # def create_option(self, name, value, label, selected, index,
        # subindex=None, attrs=None):
        value = kwargs.get('value', args[1])

        if value in self.disabled_choices:
            result['attrs'].update({'disabled': 'disabled'})

        return result


class NamedMultiWidget(forms.widgets.Widget):
    template_name = 'django/forms/widgets/multiwidget.html'

    def __init__(self, widgets, attrs=None):
        self.widgets = OrderedDict()
        for name, widget in widgets.items():
            self.widgets[name] = widget() if isinstance(widget, type) else widget

        super(NamedMultiWidget, self).__init__(attrs)

    def _get_media(self):
        "Media for a multiwidget is the combination of all media of the subwidgets"
        media = forms.widgets.Media()
        for name, widget in self.widgets.items():
            media = media + widget.media
        return media
    media = property(_get_media)

    @property
    def is_hidden(self):
        return all(widget.is_hidden for name, widget in self.widgets.items())

    def get_context(self, name, value, attrs):
        context = super(NamedMultiWidget, self).get_context(name, value, attrs)
        if self.is_localized:
            for widget in self.widgets:
                widget.is_localized = self.is_localized

        value = self.decompress(value)

        final_attrs = context['widget']['attrs']
        input_type = final_attrs.pop('type', None)
        id_ = final_attrs.get('id')
        subwidgets = []
        for widget_name, widget in self.widgets.items():
            if input_type is not None:
                widget.input_type = input_type
            full_widget_name = '%s_%s' % (name, widget_name)
            try:
                widget_value = value[widget_name]
            except IndexError:
                widget_value = None
            if id_:
                widget_attrs = final_attrs.copy()
                widget_attrs['id'] = '%s_%s' % (id_, widget_name)
            else:
                widget_attrs = final_attrs
            subwidgets.append(
                widget.get_context(
                    full_widget_name, widget_value, widget_attrs
                )['widget']
            )
        context['widget']['subwidgets'] = subwidgets
        return context

    def id_for_label(self, id_):
        if id_:
            id_ += '_{}'.format(self.widgets.keys()[0])
        return id_

    def value_from_datadict(self, data, files, name):
        return {
            name: widget.value_from_datadict(
                data, files, name + '_%s' % name
            ) for name, widget in self.widgets.items()
        }

    def value_omitted_from_data(self, data, files, name):
        return all(
            widget.value_omitted_from_data(data, files, name + '_%s' % name)
            for name, widget in self.widgets.items()
        )

    @property
    def needs_multipart_form(self):
        return any(
            widget.needs_multipart_form for name, widget in self.widgets.items()
        )


class PlainWidget(forms.widgets.Widget):
    """
    Class to define a form widget that effectively nulls the htmls of a
    widget and reduces the output to only it's value
    """
    def render(self, name, value, attrs=None):
        return mark_safe(s='%s' % value)


class TextAreaDiv(forms.widgets.Widget):
    """
    Class to define a form widget that simulates the behavior of a
    Textarea widget but using a div tag instead
    """
    template_name = 'appearance/forms/widgets/textareadiv.html'

    def __init__(self, attrs=None):
        # The 'rows' and 'cols' attributes are required for HTML correctness.
        default_attrs = {'class': 'text_area_div'}
        if attrs:
            default_attrs.update(attrs)
        super(TextAreaDiv, self).__init__(default_attrs)



from __future__ import unicode_literals

from django.utils.encoding import force_text

from ..permissions import permission_document_view

from .base import GenericDocumentViewTestCase


class DocumentPageViewTestCase(GenericDocumentViewTestCase):
    def _request_test_document_page_list_view(self):
        return self.get(
            viewname='documents:document_pages', kwargs={
                'pk': self.test_document.pk
            }
        )

    def test_document_page_list_view_no_permission(self):
        response = self._request_test_document_page_list_view()
        self.assertEqual(response.status_code, 404)

    def test_document_page_list_view_with_access(self):
        self.grant_access(
            obj=self.test_document, permission=permission_document_view
        )

        response = self._request_test_document_page_list_view()
        self.assertContains(
            response=response, text=self.test_document.label, status_code=200
        )

    def _request_test_document_page_view(self, document_page):
        return self.get(
            viewname='documents:document_page_view', kwargs={
                'pk': document_page.pk,
            }
        )

    def test_document_page_view_no_permissions(self):
        response = self._request_test_document_page_view(
            document_page=self.test_document.pages.first()
        )
        self.assertEqual(response.status_code, 404)

    def test_document_page_view_with_access(self):
        self.grant_access(
            obj=self.test_document, permission=permission_document_view
        )

        response = self._request_test_document_page_view(
            document_page=self.test_document.pages.first()
        )
        self.assertContains(
            response=response, text=force_text(
                self.test_document.pages.first()
            ), status_code=200
        )



from __future__ import unicode_literals

from django.utils.six.moves import xmlrpc_client

import mayan

from .exceptions import NotLatestVersion, UnknownLatestVersion
from .literals import MAYAN_PYPI_NAME, PYPI_URL


def check_version():
    pypi = xmlrpc_client.ServerProxy(PYPI_URL)
    versions = pypi.package_releases(MAYAN_PYPI_NAME)
    if not versions:
        raise UnknownLatestVersion
    else:
        if versions[0] != mayan.__version__:
            raise NotLatestVersion(upstream_version=versions[0])



from __future__ import unicode_literals

from actstream.models import Action

from mayan.apps.common.tests import GenericViewTestCase

from ..events import event_workflow_created, event_workflow_edited
from ..models import Workflow
from ..permissions import permission_workflow_create, permission_workflow_edit

from .mixins import WorkflowTestMixin, WorkflowViewTestMixin


class WorkflowEventsTestCase(WorkflowTestMixin, WorkflowViewTestMixin, GenericViewTestCase):
    def test_workflow_create_event_no_permissions(self):
        action_count = Action.objects.count()

        response = self._request_test_workflow_create_view()
        self.assertEqual(response.status_code, 403)

        self.assertEqual(Action.objects.count(), action_count)

    def test_workflow_create_event_with_permissions(self):
        self.grant_permission(permission=permission_workflow_create)

        action_count = Action.objects.count()

        response = self._request_test_workflow_create_view()
        self.assertEqual(response.status_code, 302)

        self.assertEqual(Action.objects.count(), action_count + 1)

        event = Action.objects.first()

        workflow = Workflow.objects.first()

        self.assertEqual(event.verb, event_workflow_created.id)
        self.assertEqual(event.target, workflow)
        self.assertEqual(event.actor, self._test_case_user)

    def test_workflow_edit_event_no_permissions(self):
        self._create_test_workflow()

        action_count = Action.objects.count()

        response = self._request_test_workflow_edit_view()
        self.assertEqual(response.status_code, 404)

        self.assertEqual(Action.objects.count(), action_count)

    def test_workflow_edit_event_with_access(self):
        self._create_test_workflow()

        self.grant_access(
            obj=self.test_workflow, permission=permission_workflow_edit
        )

        action_count = Action.objects.count()

        response = self._request_test_workflow_edit_view()
        self.assertEqual(response.status_code, 302)

        self.assertEqual(Action.objects.count(), action_count + 1)

        event = Action.objects.first()

        self.assertEqual(event.verb, event_workflow_edited.id)
        self.assertEqual(event.target, self.test_workflow)
        self.assertEqual(event.actor, self._test_case_user)



from __future__ import absolute_import, unicode_literals

from django.contrib import messages
from django.core.files.base import ContentFile
from django.db import transaction
from django.http import Http404, HttpResponseRedirect
from django.shortcuts import get_object_or_404
from django.template import RequestContext
from django.urls import reverse, reverse_lazy
from django.utils.translation import ugettext_lazy as _

from mayan.apps.common.generics import (
    AddRemoveView, ConfirmView, FormView, SingleObjectCreateView,
    SingleObjectDeleteView, SingleObjectDetailView,
    SingleObjectDynamicFormCreateView, SingleObjectDynamicFormEditView,
    SingleObjectDownloadView, SingleObjectEditView, SingleObjectListView
)
from mayan.apps.common.mixins import ExternalObjectMixin
from mayan.apps.documents.events import event_document_type_edited
from mayan.apps.documents.models import DocumentType
from mayan.apps.documents.permissions import permission_document_type_edit
from mayan.apps.events.classes import EventType
from mayan.apps.events.models import StoredEventType

from ..classes import WorkflowAction
from ..events import event_workflow_edited
from ..forms import (
    WorkflowActionSelectionForm, WorkflowForm, WorkflowPreviewForm,
    WorkflowStateActionDynamicForm, WorkflowStateForm, WorkflowTransitionForm,
    WorkflowTransitionTriggerEventRelationshipFormSet
)
from ..icons import (
    icon_workflow_list, icon_workflow_state, icon_workflow_state_action,
    icon_workflow_transition
)
from ..links import (
    link_setup_workflow_create, link_setup_workflow_state_create,
    link_setup_workflow_state_action_selection,
    link_setup_workflow_transition_create
)
from ..models import (
    Workflow, WorkflowState, WorkflowStateAction, WorkflowTransition
)
from ..permissions import (
    permission_workflow_create, permission_workflow_delete,
    permission_workflow_edit, permission_workflow_tools,
    permission_workflow_view,
)
from ..tasks import task_launch_all_workflows

__all__ = (
    'WorkflowImageView', 'WorkflowPreviewView',
    'SetupWorkflowListView', 'SetupWorkflowCreateView', 'SetupWorkflowEditView',
    'SetupWorkflowDeleteView', 'SetupWorkflowDocumentTypesView',
    'SetupWorkflowStateActionCreateView', 'SetupWorkflowStateActionDeleteView',
    'SetupWorkflowStateActionEditView', 'SetupWorkflowStateActionListView',
    'SetupWorkflowStateActionSelectionView', 'SetupWorkflowStateCreateView',
    'SetupWorkflowStateDeleteView', 'SetupWorkflowStateEditView',
    'SetupWorkflowStateListView', 'SetupWorkflowTransitionCreateView',
    'SetupWorkflowTransitionDeleteView', 'SetupWorkflowTransitionEditView',
    'SetupWorkflowTransitionListView',
    'SetupWorkflowTransitionTriggerEventListView', 'ToolLaunchAllWorkflows'
)


class SetupDocumentTypeWorkflowsView(AddRemoveView):
    main_object_permission = permission_document_type_edit
    main_object_model = DocumentType
    main_object_pk_url_kwarg = 'pk'
    secondary_object_model = Workflow
    secondary_object_permission = permission_workflow_edit
    list_available_title = _('Available workflows')
    list_added_title = _('Workflows assigned this document type')
    related_field = 'workflows'

    def get_actions_extra_kwargs(self):
        return {'_user': self.request.user}

    def get_extra_context(self):
        return {
            'object': self.main_object,
            'subtitle': _(
                'Removing a workflow from a document type will also '
                'remove all running instances of that workflow.'
            ),
            'title': _(
                'Workflows assigned the document type: %s'
            ) % self.main_object,
        }

    def action_add(self, queryset, _user):
        with transaction.atomic():
            event_document_type_edited.commit(
                actor=_user, target=self.main_object
            )

            for obj in queryset:
                self.main_object.workflows.add(obj)
                event_workflow_edited.commit(
                    action_object=self.main_object, actor=_user, target=obj
                )

    def action_remove(self, queryset, _user):
        with transaction.atomic():
            event_document_type_edited.commit(
                actor=_user, target=self.main_object
            )

            for obj in queryset:
                self.main_object.workflows.remove(obj)
                event_workflow_edited.commit(
                    action_object=self.main_object, actor=_user,
                    target=obj
                )
                obj.instances.filter(
                    document__document_type=self.main_object
                ).delete()


class SetupWorkflowListView(SingleObjectListView):
    model = Workflow
    object_permission = permission_workflow_view

    def get_extra_context(self):
        return {
            'hide_object': True,
            'no_results_icon': icon_workflow_list,
            'no_results_main_link': link_setup_workflow_create.resolve(
                context=RequestContext(request=self.request)
            ),
            'no_results_text': _(
                'Workflows store a series of states and keep track of the '
                'current state of a document. Transitions are used to change the '
                'current state to a new one.'
            ),
            'no_results_title': _(
                'No workflows have been defined'
            ),
            'title': _('Workflows'),
        }


class SetupWorkflowCreateView(SingleObjectCreateView):
    extra_context = {'title': _('Create workflow')}
    form_class = WorkflowForm
    model = Workflow
    post_action_redirect = reverse_lazy(
        viewname='document_states:setup_workflow_list'
    )
    view_permission = permission_workflow_create

    def get_save_extra_data(self):
        return {'_user': self.request.user}


class SetupWorkflowDeleteView(SingleObjectDeleteView):
    model = Workflow
    object_permission = permission_workflow_delete
    post_action_redirect = reverse_lazy(
        viewname='document_states:setup_workflow_list'
    )

    def get_extra_context(self):
        return {
            'title': _(
                'Delete workflow: %s?'
            ) % self.object,
        }


class SetupWorkflowEditView(SingleObjectEditView):
    form_class = WorkflowForm
    model = Workflow
    object_permission = permission_workflow_edit
    post_action_redirect = reverse_lazy(
        viewname='document_states:setup_workflow_list'
    )

    def get_extra_context(self):
        return {
            'title': _(
                'Edit workflow: %s'
            ) % self.object,
        }

    def get_save_extra_data(self):
        return {'_user': self.request.user}


class SetupWorkflowDocumentTypesView(AddRemoveView):
    main_object_permission = permission_workflow_edit
    main_object_model = Workflow
    main_object_pk_url_kwarg = 'pk'
    secondary_object_model = DocumentType
    secondary_object_permission = permission_document_type_edit
    list_available_title = _('Available document types')
    list_added_title = _('Document types assigned this workflow')
    related_field = 'document_types'

    def get_actions_extra_kwargs(self):
        return {'_user': self.request.user}

    def get_extra_context(self):
        return {
            'object': self.main_object,
            'subtitle': _(
                'Removing a document type from a workflow will also '
                'remove all running instances of that workflow for '
                'documents of the document type just removed.'
            ),
            'title': _(
                'Document types assigned the workflow: %s'
            ) % self.main_object,
        }

    def action_add(self, queryset, _user):
        with transaction.atomic():
            event_workflow_edited.commit(
                actor=_user, target=self.main_object
            )

            for obj in queryset:
                self.main_object.document_types.add(obj)
                event_document_type_edited.commit(
                    action_object=self.main_object, actor=_user, target=obj
                )

    def action_remove(self, queryset, _user):
        with transaction.atomic():
            event_workflow_edited.commit(
                actor=_user, target=self.main_object
            )

            for obj in queryset:
                self.main_object.document_types.remove(obj)
                event_document_type_edited.commit(
                    action_object=self.main_object, actor=_user,
                    target=obj
                )
                self.main_object.instances.filter(
                    document__document_type=obj
                ).delete()


# Workflow state actions


class SetupWorkflowStateActionCreateView(SingleObjectDynamicFormCreateView):
    form_class = WorkflowStateActionDynamicForm
    object_permission = permission_workflow_edit

    def get_class(self):
        try:
            return WorkflowAction.get(name=self.kwargs['class_path'])
        except KeyError:
            raise Http404(
                '{} class not found'.format(self.kwargs['class_path'])
            )

    def get_extra_context(self):
        return {
            'navigation_object_list': ('object', 'workflow'),
            'object': self.get_object(),
            'title': _(
                'Create a "%s" workflow action'
            ) % self.get_class().label,
            'workflow': self.get_object().workflow
        }

    def get_form_extra_kwargs(self):
        return {
            'request': self.request,
            'action_path': self.kwargs['class_path']
        }

    def get_form_schema(self):
        return self.get_class()().get_form_schema(request=self.request)

    def get_instance_extra_data(self):
        return {
            'action_path': self.kwargs['class_path'],
            'state': self.get_object()
        }

    def get_object(self):
        return get_object_or_404(klass=WorkflowState, pk=self.kwargs['pk'])

    def get_post_action_redirect(self):
        return reverse(
            viewname='document_states:setup_workflow_state_action_list',
            kwargs={'pk': self.get_object().pk}
        )


class SetupWorkflowStateActionDeleteView(SingleObjectDeleteView):
    model = WorkflowStateAction
    object_permission = permission_workflow_edit

    def get_extra_context(self):
        return {
            'navigation_object_list': (
                'object', 'workflow_state', 'workflow'
            ),
            'object': self.get_object(),
            'title': _('Delete workflow state action: %s') % self.get_object(),
            'workflow': self.get_object().state.workflow,
            'workflow_state': self.get_object().state,
        }

    def get_post_action_redirect(self):
        return reverse(
            viewname='document_states:setup_workflow_state_action_list',
            kwargs={'pk': self.get_object().state.pk}
        )


class SetupWorkflowStateActionEditView(SingleObjectDynamicFormEditView):
    form_class = WorkflowStateActionDynamicForm
    model = WorkflowStateAction
    object_permission = permission_workflow_edit

    def get_extra_context(self):
        return {
            'navigation_object_list': (
                'object', 'workflow_state', 'workflow'
            ),
            'object': self.get_object(),
            'title': _('Edit workflow state action: %s') % self.get_object(),
            'workflow': self.get_object().state.workflow,
            'workflow_state': self.get_object().state,
        }

    def get_form_extra_kwargs(self):
        return {
            'request': self.request,
            'action_path': self.get_object().action_path,
        }

    def get_form_schema(self):
        return self.get_object().get_class_instance().get_form_schema(
            request=self.request
        )

    def get_post_action_redirect(self):
        return reverse(
            viewname='document_states:setup_workflow_state_action_list',
            kwargs={'pk': self.get_object().state.pk}
        )


class SetupWorkflowStateActionListView(SingleObjectListView):
    object_permission = permission_workflow_edit

    def get_extra_context(self):
        return {
            'hide_object': True,
            'navigation_object_list': ('object', 'workflow'),
            'no_results_icon': icon_workflow_state_action,
            'no_results_main_link': link_setup_workflow_state_action_selection.resolve(
                context=RequestContext(
                    request=self.request, dict_={
                        'object': self.get_workflow_state()
                    }
                )
            ),
            'no_results_text': _(
                'Workflow state actions are macros that get executed when '
                'documents enters or leaves the state in which they reside.'
            ),
            'no_results_title': _(
                'There are no actions for this workflow state'
            ),
            'object': self.get_workflow_state(),
            'title': _(
                'Actions for workflow state: %s'
            ) % self.get_workflow_state(),
            'workflow': self.get_workflow_state().workflow,
        }

    def get_form_schema(self):
        return {'fields': self.get_class().fields}

    def get_source_queryset(self):
        return self.get_workflow_state().actions.all()

    def get_workflow_state(self):
        return get_object_or_404(klass=WorkflowState, pk=self.kwargs['pk'])


class SetupWorkflowStateActionSelectionView(FormView):
    form_class = WorkflowActionSelectionForm
    view_permission = permission_workflow_edit

    def form_valid(self, form):
        klass = form.cleaned_data['klass']
        return HttpResponseRedirect(
            redirect_to=reverse(
                viewname='document_states:setup_workflow_state_action_create',
                kwargs={'pk': self.get_object().pk, 'class_path': klass}
            )
        )

    def get_extra_context(self):
        return {
            'navigation_object_list': (
                'object', 'workflow'
            ),
            'object': self.get_object(),
            'title': _('New workflow state action selection'),
            'workflow': self.get_object().workflow,
        }

    def get_object(self):
        return get_object_or_404(klass=WorkflowState, pk=self.kwargs['pk'])


# Workflow states


class SetupWorkflowStateCreateView(ExternalObjectMixin, SingleObjectCreateView):
    external_object_class = Workflow
    external_object_permission = permission_workflow_edit
    external_object_pk_url_kwarg = 'pk'
    form_class = WorkflowStateForm

    def get_extra_context(self):
        return {
            'object': self.get_workflow(),
            'title': _(
                'Create states for workflow: %s'
            ) % self.get_workflow()
        }

    def get_instance_extra_data(self):
        return {'workflow': self.get_workflow()}

    def get_source_queryset(self):
        return self.get_workflow().states.all()

    def get_success_url(self):
        return reverse(
            viewname='document_states:setup_workflow_state_list',
            kwargs={'pk': self.kwargs['pk']}
        )

    def get_workflow(self):
        return self.external_object


class SetupWorkflowStateDeleteView(SingleObjectDeleteView):
    model = WorkflowState
    object_permission = permission_workflow_edit
    pk_url_kwarg = 'pk'

    def get_extra_context(self):
        return {
            'navigation_object_list': ('object', 'workflow_instance'),
            'object': self.get_object(),
            'title': _(
                'Delete workflow state: %s?'
            ) % self.object,
            'workflow_instance': self.get_object().workflow,
        }

    def get_success_url(self):
        return reverse(
            viewname='document_states:setup_workflow_state_list',
            kwargs={'pk': self.get_object().workflow.pk}
        )


class SetupWorkflowStateEditView(SingleObjectEditView):
    form_class = WorkflowStateForm
    model = WorkflowState
    object_permission = permission_workflow_edit
    pk_url_kwarg = 'pk'

    def get_extra_context(self):
        return {
            'navigation_object_list': ('object', 'workflow_instance'),
            'object': self.get_object(),
            'title': _(
                'Edit workflow state: %s'
            ) % self.object,
            'workflow_instance': self.get_object().workflow,
        }

    def get_success_url(self):
        return reverse(
            viewname='document_states:setup_workflow_state_list',
            kwargs={'pk': self.get_object().workflow.pk}
        )


class SetupWorkflowStateListView(ExternalObjectMixin, SingleObjectListView):
    external_object_class = Workflow
    external_object_permission = permission_workflow_view
    external_object_pk_url_kwarg = 'pk'
    object_permission = permission_workflow_view

    def get_extra_context(self):
        return {
            'hide_object': True,
            'no_results_icon': icon_workflow_state,
            'no_results_main_link': link_setup_workflow_state_create.resolve(
                context=RequestContext(
                    self.request, {'object': self.get_workflow()}
                )
            ),
            'no_results_text': _(
                'Create states and link them using transitions.'
            ),
            'no_results_title': _(
                'This workflow doesn\'t have any states'
            ),
            'object': self.get_workflow(),
            'title': _('States of workflow: %s') % self.get_workflow()
        }

    def get_source_queryset(self):
        return self.get_workflow().states.all()

    def get_workflow(self):
        return self.external_object


# Transitions


class SetupWorkflowTransitionCreateView(ExternalObjectMixin, SingleObjectCreateView):
    external_object_class = Workflow
    external_object_permission = permission_workflow_edit
    external_object_pk_url_kwarg = 'pk'
    form_class = WorkflowTransitionForm

    def get_extra_context(self):
        return {
            'object': self.get_workflow(),
            'title': _(
                'Create transitions for workflow: %s'
            ) % self.get_workflow()
        }

    def get_form_kwargs(self):
        kwargs = super(
            SetupWorkflowTransitionCreateView, self
        ).get_form_kwargs()
        kwargs['workflow'] = self.get_workflow()
        return kwargs

    def get_instance_extra_data(self):
        return {'workflow': self.get_workflow()}

    def get_source_queryset(self):
        return self.get_workflow().transitions.all()

    def get_success_url(self):
        return reverse(
            viewname='document_states:setup_workflow_transition_list',
            kwargs={'pk': self.kwargs['pk']}
        )

    def get_workflow(self):
        return self.external_object


class SetupWorkflowTransitionDeleteView(SingleObjectDeleteView):
    model = WorkflowTransition
    object_permission = permission_workflow_edit
    pk_url_kwarg = 'pk'

    def get_extra_context(self):
        return {
            'object': self.get_object(),
            'navigation_object_list': ('object', 'workflow_instance'),
            'title': _(
                'Delete workflow transition: %s?'
            ) % self.object,
            'workflow_instance': self.get_object().workflow,
        }

    def get_success_url(self):
        return reverse(
            viewname='document_states:setup_workflow_transition_list',
            kwargs={'pk': self.get_object().workflow.pk}
        )


class SetupWorkflowTransitionEditView(SingleObjectEditView):
    form_class = WorkflowTransitionForm
    model = WorkflowTransition
    object_permission = permission_workflow_edit
    pk_url_kwarg = 'pk'

    def get_extra_context(self):
        return {
            'navigation_object_list': ('object', 'workflow_instance'),
            'object': self.get_object(),
            'title': _(
                'Edit workflow transition: %s'
            ) % self.object,
            'workflow_instance': self.get_object().workflow,
        }

    def get_form_kwargs(self):
        kwargs = super(
            SetupWorkflowTransitionEditView, self
        ).get_form_kwargs()
        kwargs['workflow'] = self.get_object().workflow
        return kwargs

    def get_success_url(self):
        return reverse(
            viewname='document_states:setup_workflow_transition_list',
            kwargs={'pk': self.get_object().workflow.pk}
        )


class SetupWorkflowTransitionListView(ExternalObjectMixin, SingleObjectListView):
    external_object_class = Workflow
    external_object_permission = permission_workflow_view
    external_object_pk_url_kwarg = 'pk'
    object_permission = permission_workflow_view

    def get_extra_context(self):
        return {
            'hide_object': True,
            'no_results_icon': icon_workflow_transition,
            'no_results_main_link': link_setup_workflow_transition_create.resolve(
                context=RequestContext(
                    self.request, {'object': self.get_workflow()}
                )
            ),
            'no_results_text': _(
                'Create a transition and use it to move a workflow from '
                ' one state to another.'
            ),
            'no_results_title': _(
                'This workflow doesn\'t have any transitions'
            ),
            'object': self.get_workflow(),
            'title': _(
                'Transitions of workflow: %s'
            ) % self.get_workflow()
        }

    def get_source_queryset(self):
        return self.get_workflow().transitions.all()

    def get_workflow(self):
        return self.external_object


class SetupWorkflowTransitionTriggerEventListView(ExternalObjectMixin, FormView):
    external_object_class = WorkflowTransition
    external_object_permission = permission_workflow_edit
    external_object_pk_url_kwarg = 'pk'
    form_class = WorkflowTransitionTriggerEventRelationshipFormSet

    def dispatch(self, *args, **kwargs):
        EventType.refresh()
        return super(
            SetupWorkflowTransitionTriggerEventListView, self
        ).dispatch(*args, **kwargs)

    def form_valid(self, form):
        try:
            for instance in form:
                instance.save()
        except Exception as exception:
            messages.error(
                message=_(
                    'Error updating workflow transition trigger events; %s'
                ) % exception, request=self.request

            )
        else:
            messages.success(
                message=_(
                    'Workflow transition trigger events updated successfully'
                ), request=self.request
            )

        return super(
            SetupWorkflowTransitionTriggerEventListView, self
        ).form_valid(form=form)

    def get_extra_context(self):
        return {
            'form_display_mode_table': True,
            'navigation_object_list': ('object', 'workflow'),
            'object': self.get_object(),
            'subtitle': _(
                'Triggers are events that cause this transition to execute '
                'automatically.'
            ),
            'title': _(
                'Workflow transition trigger events for: %s'
            ) % self.get_object(),
            'workflow': self.get_object().workflow,
        }

    def get_initial(self):
        obj = self.get_object()
        initial = []

        # Return the queryset by name from the sorted list of the class
        event_type_ids = [event_type.id for event_type in EventType.all()]
        event_type_queryset = StoredEventType.objects.filter(
            name__in=event_type_ids
        )

        # Sort queryset in Python by namespace, then by label
        event_type_queryset = sorted(
            event_type_queryset, key=lambda x: (x.namespace, x.label)
        )

        for event_type in event_type_queryset:
            initial.append({
                'transition': obj,
                'event_type': event_type,
            })
        return initial

    def get_object(self):
        return self.external_object

    def get_post_action_redirect(self):
        return reverse(
            viewname='document_states:setup_workflow_transition_list',
            kwargs={'pk': self.get_object().workflow.pk}
        )


class ToolLaunchAllWorkflows(ConfirmView):
    extra_context = {
        'title': _('Launch all workflows?'),
        'subtitle': _(
            'This will launch all workflows created after documents have '
            'already been uploaded.'
        )
    }
    view_permission = permission_workflow_tools

    def view_action(self):
        task_launch_all_workflows.apply_async()
        messages.success(
            message=_('Workflow launch queued successfully.'),
            request=self.request
        )


class WorkflowImageView(SingleObjectDownloadView):
    attachment = False
    model = Workflow
    object_permission = permission_workflow_view

    def get_file(self):
        workflow = self.get_object()
        return ContentFile(workflow.render(), name=workflow.label)

    def get_mimetype(self):
        return 'image'


class WorkflowPreviewView(SingleObjectDetailView):
    form_class = WorkflowPreviewForm
    model = Workflow
    object_permission = permission_workflow_view

    def get_extra_context(self):
        return {
            'hide_labels': True,
            'title': _('Preview of: %s') % self.get_object()
        }



from __future__ import unicode_literals

import types

from django.utils import six

if six.PY3:
    dict_type = dict
    dictionary_type = dict
else:
    dict_type = types.DictType
    dictionary_type = types.DictionaryType

try:
    from email.Utils import collapse_rfc2231_value  # NOQA
except ImportError:
    from email.utils import collapse_rfc2231_value  # NOQA

try:
    FileNotFoundError
except NameError:
    FileNotFoundErrorException = IOError
else:
    FileNotFoundErrorException = FileNotFoundError  # NOQA



from __future__ import unicode_literals

import glob
import importlib
import logging
import os
import random

from furl import furl

from django.apps import apps
from django.conf import settings
from django.conf.urls import url
from django.contrib.contenttypes.models import ContentType
from django.db import connection, connections, models
from django.db.models.signals import post_save, pre_save
from django.http import HttpResponse
from django.template import Context, Template
from django.test.utils import ContextList
from django.urls import clear_url_caches, reverse
from django.utils.encoding import force_bytes
from django.utils.six import PY3

from mayan.apps.storage.settings import setting_temporary_directory

from .literals import TEST_VIEW_NAME, TEST_VIEW_URL


if getattr(settings, 'COMMON_TEST_FILE_HANDLES', False):
    import psutil


class ClientMethodsTestCaseMixin(object):
    def _build_verb_kwargs(self, viewname=None, path=None, *args, **kwargs):
        data = kwargs.pop('data', {})
        follow = kwargs.pop('follow', False)
        query = kwargs.pop('query', {})

        if viewname:
            path = reverse(viewname=viewname, *args, **kwargs)

        path = furl(url=path)
        path.args.update(query)

        return {'follow': follow, 'data': data, 'path': path.tostr()}

    def delete(self, viewname=None, path=None, *args, **kwargs):
        return self.client.delete(
            **self._build_verb_kwargs(
                path=path, viewname=viewname, *args, **kwargs
            )
        )

    def get(self, viewname=None, path=None, *args, **kwargs):
        return self.client.get(
            **self._build_verb_kwargs(
                path=path, viewname=viewname, *args, **kwargs
            )
        )

    def patch(self, viewname=None, path=None, *args, **kwargs):
        return self.client.patch(
            **self._build_verb_kwargs(
                path=path, viewname=viewname, *args, **kwargs
            )
        )

    def post(self, viewname=None, path=None, *args, **kwargs):
        return self.client.post(
            **self._build_verb_kwargs(
                path=path, viewname=viewname, *args, **kwargs
            )
        )

    def put(self, viewname=None, path=None, *args, **kwargs):
        return self.client.put(
            **self._build_verb_kwargs(
                path=path, viewname=viewname, *args, **kwargs
            )
        )


class ConnectionsCheckTestCaseMixin(object):
    _open_connections_check_enable = True

    def _get_open_connections_count(self):
        return len(connections.all())

    def setUp(self):
        super(ConnectionsCheckTestCaseMixin, self).setUp()
        self._connections_count = self._get_open_connections_count()

    def tearDown(self):
        if self._open_connections_check_enable:

            self.assertEqual(
                self._connections_count, self._get_open_connections_count(),
                msg='Database connection leak. The number of database '
                'connections at the start and at the end of the test are not '
                'the same.'
            )

        super(ConnectionsCheckTestCaseMixin, self).tearDown()


class ContentTypeCheckTestCaseMixin(object):
    expected_content_type = 'text/html; charset=utf-8'

    def _pre_setup(self):
        super(ContentTypeCheckTestCaseMixin, self)._pre_setup()
        test_instance = self

        class CustomClient(self.client_class):
            def request(self, *args, **kwargs):
                response = super(CustomClient, self).request(*args, **kwargs)

                content_type = response._headers.get('content-type', [None, ''])[1]
                if test_instance.expected_content_type:
                    test_instance.assertEqual(
                        content_type, test_instance.expected_content_type,
                        msg='Unexpected response content type: {}, expected: {}.'.format(
                            content_type, test_instance.expected_content_type
                        )
                    )

                return response

        self.client = CustomClient()


class ModelTestCaseMixin(object):
    def _model_instance_to_dictionary(self, instance):
        return instance._meta.model._default_manager.filter(
            pk=instance.pk
        ).values()[0]


class OpenFileCheckTestCaseMixin(object):
    def _get_descriptor_count(self):
        process = psutil.Process()
        return process.num_fds()

    def _get_open_files(self):
        process = psutil.Process()
        return process.open_files()

    def setUp(self):
        super(OpenFileCheckTestCaseMixin, self).setUp()
        if getattr(settings, 'COMMON_TEST_FILE_HANDLES', False):
            self._open_files = self._get_open_files()

    def tearDown(self):
        if getattr(settings, 'COMMON_TEST_FILE_HANDLES', False) and not getattr(self, '_skip_file_descriptor_test', False):
            for new_open_file in self._get_open_files():
                self.assertFalse(
                    new_open_file not in self._open_files,
                    msg='File descriptor leak. The number of file descriptors '
                    'at the start and at the end of the test are not the same.'
                )

            self._skip_file_descriptor_test = False

        super(OpenFileCheckTestCaseMixin, self).tearDown()


class RandomPrimaryKeyModelMonkeyPatchMixin(object):
    random_primary_key_random_floor = 100
    random_primary_key_random_ceiling = 10000
    random_primary_key_maximum_attempts = 100
    random_primary_key_enable = True

    @staticmethod
    def get_unique_primary_key(model):
        pk_list = model._meta.default_manager.values_list('pk', flat=True)

        attempts = 0
        while True:
            primary_key = random.randint(
                RandomPrimaryKeyModelMonkeyPatchMixin.random_primary_key_random_floor,
                RandomPrimaryKeyModelMonkeyPatchMixin.random_primary_key_random_ceiling
            )

            if primary_key not in pk_list:
                break

            attempts = attempts + 1

            if attempts > RandomPrimaryKeyModelMonkeyPatchMixin.random_primary_key_maximum_attempts:
                raise Exception(
                    'Maximum number of retries for an unique random primary '
                    'key reached.'
                )

        return primary_key

    def setUp(self):
        if self.random_primary_key_enable:
            self.method_save_original = models.Model.save

            def method_save_new(instance, *args, **kwargs):
                if instance.pk:
                    return self.method_save_original(instance, *args, **kwargs)
                else:
                    # Set meta.auto_created to True to have the original save_base
                    # not send the pre_save signal which would normally send
                    # the instance without a primary key. Since we assign a random
                    # primary key any pre_save signal handler that relies on an
                    # empty primary key will fail.
                    # The meta.auto_created and manual pre_save sending emulates
                    # the original behavior. Since meta.auto_created also disables
                    # the post_save signal we must also send it ourselves.
                    # This hack work with Django 1.11 .save_base() but can break
                    # in future versions if that method is updated.
                    pre_save.send(
                        sender=instance.__class__, instance=instance, raw=False,
                        update_fields=None,
                    )
                    instance._meta.auto_created = True
                    instance.pk = RandomPrimaryKeyModelMonkeyPatchMixin.get_unique_primary_key(
                        model=instance._meta.model
                    )
                    instance.id = instance.pk

                    result = instance.save_base(force_insert=True)
                    instance._meta.auto_created = False

                    post_save.send(
                        sender=instance.__class__, instance=instance, created=True,
                        update_fields=None, raw=False
                    )

                    return result

            setattr(models.Model, 'save', method_save_new)
        super(RandomPrimaryKeyModelMonkeyPatchMixin, self).setUp()

    def tearDown(self):
        if self.random_primary_key_enable:
            models.Model.save = self.method_save_original
        super(RandomPrimaryKeyModelMonkeyPatchMixin, self).tearDown()


class SilenceLoggerTestCaseMixin(object):
    """
    Changes the log level of a specific logger for the duration of a test.
    The default level for silenced loggers is CRITICAL.
    Example: self._silence_logger(name='mayan.apps.converter.managers')
    """
    test_case_silenced_logger = None
    test_case_silenced_logger_new_level = logging.CRITICAL

    def tearDown(self):
        if self.test_case_silenced_logger:
            self.test_case_silenced_logger.setLevel(
                level=self.test_case_silenced_logger_level
            )

        super(SilenceLoggerTestCaseMixin, self).tearDown()

    def _silence_logger(self, name):
        self.test_case_silenced_logger = logging.getLogger(name=name)
        self.test_case_silenced_logger_level = self.test_case_silenced_logger.level
        self.test_case_silenced_logger.setLevel(
            level=self.test_case_silenced_logger_new_level
        )


class TempfileCheckTestCasekMixin(object):
    # Ignore the jvmstat instrumentation and GitLab's CI .config files
    # Ignore LibreOffice fontconfig cache dir
    ignore_globs = ('hsperfdata_*', '.config', '.cache')

    def _get_temporary_entries(self):
        ignored_result = []

        # Expand globs by joining the temporary directory and then flattening
        # the list of lists into a single list
        for item in self.ignore_globs:
            ignored_result.extend(
                glob.glob(
                    os.path.join(setting_temporary_directory.value, item)
                )
            )

        # Remove the path and leave only the expanded filename
        ignored_result = map(lambda x: os.path.split(x)[-1], ignored_result)

        return set(
            os.listdir(setting_temporary_directory.value)
        ) - set(ignored_result)

    def setUp(self):
        super(TempfileCheckTestCasekMixin, self).setUp()
        if getattr(settings, 'COMMON_TEST_TEMP_FILES', False):
            self._temporary_items = self._get_temporary_entries()

    def tearDown(self):
        if getattr(settings, 'COMMON_TEST_TEMP_FILES', False):
            final_temporary_items = self._get_temporary_entries()
            self.assertEqual(
                self._temporary_items, final_temporary_items,
                msg='Orphan temporary file. The number of temporary files and/or '
                'directories at the start and at the end of the test are not the '
                'same. Orphan entries: {}'.format(
                    ','.join(final_temporary_items - self._temporary_items)
                )
            )
        super(TempfileCheckTestCasekMixin, self).tearDown()


class TestModelTestMixin(object):
    _test_models = []

    def _create_test_model(self, fields=None, model_name='TestModel', options=None):

        if connection.vendor == 'mysql':
            self.skipTest(
                reason='MySQL doesn\'t support schema changes inside an '
                'atomic block.'
            )

        # Obtain the app_config and app_label from the test's module path
        app_config = apps.get_containing_app_config(
            object_name=self.__class__.__module__
        )
        app_label = app_config.label

        class Meta:
            pass

        setattr(Meta, 'app_label', app_label)

        if options is not None:
            for key, value in options.items():
                setattr(Meta, key, value)

        def save(instance, *args, **kwargs):
            # Custom .save() method to use random primary key values.
            if instance.pk:
                return models.Model.self(instance, *args, **kwargs)
            else:
                instance.pk = RandomPrimaryKeyModelMonkeyPatchMixin.get_unique_primary_key(
                    model=instance._meta.model
                )
                instance.id = instance.pk

                return instance.save_base(force_insert=True)

        attrs = {
            '__module__': self.__class__.__module__, 'save': save, 'Meta': Meta
        }

        if fields:
            attrs.update(fields)

        # Clear previous model registration before re-registering it again to
        # avoid conflict with test models with the same name, in the same app
        # but from another test module.
        apps.all_models[app_label].pop(model_name.lower(), None)

        if PY3:
            TestModel = type(
                model_name, (models.Model,), attrs
            )
        else:
            TestModel = type(
                force_bytes(model_name), (models.Model,), attrs
            )

        setattr(self, model_name, TestModel)
        self._test_models.append(TestModel)

        with connection.schema_editor() as schema_editor:
            schema_editor.create_model(model=TestModel)

        ContentType.objects.clear_cache()

    def _create_test_object(self, model_name='TestModel', **kwargs):
        TestModel = getattr(self, model_name)

        self.test_object = TestModel.objects.create(**kwargs)


class TestViewTestCaseMixin(object):
    has_test_view = False

    def tearDown(self):
        urlconf = importlib.import_module(settings.ROOT_URLCONF)

        self.client.logout()
        if self.has_test_view:
            urlconf.urlpatterns.pop(0)
        super(TestViewTestCaseMixin, self).tearDown()

    def add_test_view(self, test_object):
        urlconf = importlib.import_module(settings.ROOT_URLCONF)

        def test_view(request):
            template = Template('{{ object }}')
            context = Context(
                {'object': test_object, 'resolved_object': test_object}
            )
            return HttpResponse(template.render(context=context))

        urlconf.urlpatterns.insert(0, url(TEST_VIEW_URL, test_view, name=TEST_VIEW_NAME))
        clear_url_caches()
        self.has_test_view = True

    def get_test_view(self):
        response = self.get(TEST_VIEW_NAME)
        if isinstance(response.context, ContextList):
            # template widget rendering causes test client response to be
            # ContextList rather than RequestContext. Typecast to dictionary
            # before updating.
            result = dict(response.context).copy()
            result.update({'request': response.wsgi_request})
            return Context(result)
        else:
            response.context.update({'request': response.wsgi_request})
            return Context(response.context)



from __future__ import absolute_import, unicode_literals

from django.shortcuts import get_object_or_404
from django.template import RequestContext
from django.utils.translation import ugettext_lazy as _

from mayan.apps.acls.models import AccessControlList
from mayan.apps.common.generics import SingleObjectListView
from mayan.apps.documents.models import Document
from mayan.apps.documents.views import DocumentListView

from ..icons import icon_workflow_list
from ..links import link_setup_workflow_create, link_setup_workflow_state_create
from ..models import WorkflowRuntimeProxy, WorkflowStateRuntimeProxy
from ..permissions import permission_workflow_view

__all__ = (
    'WorkflowDocumentListView', 'WorkflowListView',
    'WorkflowStateDocumentListView', 'WorkflowStateListView'
)


class WorkflowDocumentListView(DocumentListView):
    def dispatch(self, request, *args, **kwargs):
        self.workflow = get_object_or_404(
            klass=WorkflowRuntimeProxy, pk=self.kwargs['pk']
        )

        AccessControlList.objects.check_access(
            obj=self.workflow, permissions=(permission_workflow_view,),
            user=request.user
        )

        return super(
            WorkflowDocumentListView, self
        ).dispatch(request, *args, **kwargs)

    def get_document_queryset(self):
        return Document.objects.filter(workflows__workflow=self.workflow)

    def get_extra_context(self):
        context = super(WorkflowDocumentListView, self).get_extra_context()
        context.update(
            {
                'no_results_text': _(
                    'Associate a workflow with some document types and '
                    'documents of those types will be listed in this view.'
                ),
                'no_results_title': _(
                    'There are no documents executing this workflow'
                ),
                'object': self.workflow,
                'title': _('Documents with the workflow: %s') % self.workflow
            }
        )
        return context


class WorkflowListView(SingleObjectListView):
    object_permission = permission_workflow_view

    def get_extra_context(self):
        return {
            'hide_object': True,
            'no_results_icon': icon_workflow_list,
            'no_results_main_link': link_setup_workflow_create.resolve(
                context=RequestContext(request=self.request)
            ),
            'no_results_text': _(
                'Create some workflows and associated them with a document '
                'type. Active workflows will be shown here and the documents '
                'for which they are executing.'
            ),
            'no_results_title': _('There are no workflows'),
            'title': _('Workflows'),
        }

    def get_source_queryset(self):
        return WorkflowRuntimeProxy.objects.all()


class WorkflowStateDocumentListView(DocumentListView):
    def get_document_queryset(self):
        return self.get_workflow_state().get_documents()

    def get_extra_context(self):
        workflow_state = self.get_workflow_state()
        context = super(WorkflowStateDocumentListView, self).get_extra_context()
        context.update(
            {
                'object': workflow_state,
                'navigation_object_list': ('object', 'workflow'),
                'no_results_title': _(
                    'There are no documents in this workflow state'
                ),
                'title': _(
                    'Documents in the workflow "%s", state "%s"'
                ) % (
                    workflow_state.workflow, workflow_state
                ),
                'workflow': WorkflowRuntimeProxy.objects.get(
                    pk=workflow_state.workflow.pk
                ),
            }
        )
        return context

    def get_workflow_state(self):
        workflow_state = get_object_or_404(
            klass=WorkflowStateRuntimeProxy, pk=self.kwargs['pk']
        )

        AccessControlList.objects.check_access(
            obj=workflow_state.workflow,
            permissions=(permission_workflow_view,), user=self.request.user
        )

        return workflow_state


class WorkflowStateListView(SingleObjectListView):
    def dispatch(self, request, *args, **kwargs):
        AccessControlList.objects.check_access(
            obj=self.get_workflow(), permissions=(permission_workflow_view,),
            user=request.user
        )

        return super(
            WorkflowStateListView, self
        ).dispatch(request, *args, **kwargs)

    def get_extra_context(self):
        return {
            'hide_link': True,
            'hide_object': True,
            'no_results_main_link': link_setup_workflow_state_create.resolve(
                context=RequestContext(
                    request=self.request, dict_={'object': self.get_workflow()}
                )
            ),
            'no_results_text': _(
                'Create states and link them using transitions.'
            ),
            'no_results_title': _(
                'This workflow doesn\'t have any state'
            ),
            'object': self.get_workflow(),
            'title': _('States of workflow: %s') % self.get_workflow()
        }

    def get_source_queryset(self):
        return WorkflowStateRuntimeProxy.objects.filter(
            workflow=self.get_workflow()
        )

    def get_workflow(self):
        return get_object_or_404(
            klass=WorkflowRuntimeProxy, pk=self.kwargs['pk']
        )



from __future__ import absolute_import, unicode_literals

import logging

from django.apps import apps
from django.db import models, transaction
from django.utils.timezone import now

from mayan.apps.documents.models import Document

from .events import (
    event_document_auto_check_in, event_document_check_in,
    event_document_forceful_check_in
)
from .exceptions import DocumentNotCheckedOut
from .literals import STATE_CHECKED_OUT, STATE_CHECKED_IN

logger = logging.getLogger(__name__)


class DocumentCheckoutManager(models.Manager):
    def are_document_new_versions_allowed(self, document, user=None):
        try:
            check_out_info = self.document_check_out_info(document=document)
        except DocumentNotCheckedOut:
            return True
        else:
            return not check_out_info.block_new_version

    def check_in_document(self, document, user=None):
        try:
            document_check_out = self.model.objects.get(document=document)
        except self.model.DoesNotExist:
            raise DocumentNotCheckedOut
        else:
            with transaction.atomic():
                if user:
                    if self.get_check_out_info(document=document).user != user:
                        event_document_forceful_check_in.commit(
                            actor=user, target=document
                        )
                    else:
                        event_document_check_in.commit(actor=user, target=document)
                else:
                    event_document_auto_check_in.commit(target=document)

                document_check_out.delete()

    def check_in_expired_check_outs(self):
        for document in self.expired_check_outs():
            document.check_in()

    def check_out_document(self, document, expiration_datetime, user, block_new_version=True):
        return self.create(
            block_new_version=block_new_version, document=document,
            expiration_datetime=expiration_datetime, user=user
        )

    def checked_out_documents(self):
        return Document.objects.filter(
            pk__in=self.model.objects.values('document__id')
        )

    def get_check_out_info(self, document):
        try:
            return self.model.objects.get(document=document)
        except self.model.DoesNotExist:
            raise DocumentNotCheckedOut

    def get_check_out_state(self, document):
        if self.is_checked_out(document=document):
            return STATE_CHECKED_OUT
        else:
            return STATE_CHECKED_IN

    def expired_check_outs(self):
        expired_list = Document.objects.filter(
            pk__in=self.model.objects.filter(
                expiration_datetime__lte=now()
            ).values_list('document__pk', flat=True)
        )
        logger.debug('expired_list: %s', expired_list)
        return expired_list

    def get_by_natural_key(self, document_natural_key):
        Document = apps.get_model(
            app_label='documents', model_name='Document'
        )
        try:
            document = Document.objects.get_by_natural_key(document_natural_key)
        except Document.DoesNotExist:
            raise self.model.DoesNotExist

        return self.get(document__pk=document.pk)

    def is_checked_out(self, document):
        return self.filter(document=document).exists()


class NewVersionBlockManager(models.Manager):
    def block(self, document):
        self.get_or_create(document=document)

    def is_blocked(self, document):
        return self.filter(document=document).exists()

    def get_by_natural_key(self, document_natural_key):
        Document = apps.get_model(
            app_label='documents', model_name='Document'
        )
        try:
            document = Document.objects.get_by_natural_key(document_natural_key)
        except Document.DoesNotExist:
            raise self.model.DoesNotExist

        return self.get(document__pk=document.pk)

    def unblock(self, document):
        self.filter(document=document).delete()



from __future__ import unicode_literals

import os

from ..models import DocumentType
from ..permissions import (
    permission_document_properties_edit,
    permission_document_type_create, permission_document_type_delete,
    permission_document_type_edit, permission_document_type_view,
)

from .base import GenericDocumentViewTestCase
from .literals import (
    TEST_DOCUMENT_TYPE_LABEL, TEST_DOCUMENT_TYPE_LABEL_EDITED,
    TEST_DOCUMENT_TYPE_QUICK_LABEL, TEST_DOCUMENT_TYPE_QUICK_LABEL_EDITED
)
from .mixins import (
    DocumentTypeQuickLabelTestMixin, DocumentTypeQuickLabelViewTestMixin,
    DocumentTypeViewTestMixin
)


class DocumentTypeViewsTestCase(
    DocumentTypeViewTestMixin, GenericDocumentViewTestCase
):
    auto_upload_document = False

    def test_document_type_create_view_no_permission(self):
        self.test_document_type.delete()

        response = self._request_test_document_type_create_view()
        self.assertEqual(response.status_code, 403)

        self.assertEqual(DocumentType.objects.count(), 0)

    def test_document_type_create_view_with_permission(self):
        self.test_document_type.delete()
        self.grant_permission(permission=permission_document_type_create)

        response = self._request_test_document_type_create_view()
        self.assertEqual(response.status_code, 302)

        self.assertEqual(DocumentType.objects.count(), 1)
        self.assertEqual(
            DocumentType.objects.first().label, TEST_DOCUMENT_TYPE_LABEL
        )

    def test_document_type_delete_view_no_permission(self):
        response = self._request_test_document_type_delete_view()
        self.assertEqual(response.status_code, 404)

        self.assertEqual(DocumentType.objects.count(), 1)

    def test_document_type_delete_view_with_access(self):
        self.grant_access(
            obj=self.test_document_type,
            permission=permission_document_type_delete
        )

        response = self._request_test_document_type_delete_view()
        self.assertEqual(response.status_code, 302)

        self.assertEqual(DocumentType.objects.count(), 0)

    def test_document_type_edit_view_no_permission(self):
        response = self._request_test_document_type_edit_view()

        self.assertEqual(response.status_code, 404)
        self.test_document_type.refresh_from_db()

        self.assertEqual(
            self.test_document_type.label, TEST_DOCUMENT_TYPE_LABEL
        )

    def test_document_type_edit_view_with_access(self):
        self.grant_access(
            obj=self.test_document_type,
            permission=permission_document_type_edit
        )

        response = self._request_test_document_type_edit_view()
        self.assertEqual(response.status_code, 302)

        self.test_document_type.refresh_from_db()
        self.assertEqual(
            self.test_document_type.label, TEST_DOCUMENT_TYPE_LABEL_EDITED
        )

    def test_document_type_list_view_no_permission(self):
        response = self._request_test_document_type_list_view()
        self.assertNotContains(
            response=response, text=self.test_document_type, status_code=200
        )

    def test_document_type_list_view_with_access(self):
        self.grant_access(
            obj=self.test_document_type,
            permission=permission_document_type_view
        )

        response = self._request_test_document_type_list_view()
        self.assertContains(
            response=response, text=self.test_document_type, status_code=200
        )


class DocumentTypeQuickLabelViewsTestCase(
    DocumentTypeQuickLabelTestMixin, DocumentTypeQuickLabelViewTestMixin,
    GenericDocumentViewTestCase
):
    auto_upload_document = False

    def test_document_type_quick_label_create_no_access(self):
        self.grant_access(
            obj=self.test_document_type,
            permission=permission_document_type_view
        )

        response = self._request_quick_label_create()
        self.assertEqual(response.status_code, 404)

        self.assertEqual(self.test_document_type.filenames.count(), 0)

    def test_document_type_quick_label_create_with_access(self):
        self.grant_access(
            obj=self.test_document_type,
            permission=permission_document_type_edit
        )

        response = self._request_quick_label_create()
        self.assertEqual(response.status_code, 302)

        self.assertEqual(self.test_document_type.filenames.count(), 1)

    def test_document_type_quick_label_delete_no_access(self):
        self._create_test_quick_label()
        response = self._request_quick_label_delete()
        self.assertEqual(response.status_code, 404)

        self.assertEqual(
            self.test_document_type.filenames.count(), 1
        )

    def test_document_type_quick_label_delete_with_access(self):
        self.grant_access(
            obj=self.test_document_type,
            permission=permission_document_type_edit
        )
        self._create_test_quick_label()

        response = self._request_quick_label_delete()
        self.assertEqual(response.status_code, 302)

        self.assertEqual(
            self.test_document_type.filenames.count(), 0
        )

    def test_document_type_quick_label_edit_no_access(self):
        self._create_test_quick_label()

        response = self._request_quick_label_edit()
        self.assertEqual(response.status_code, 404)

        self.test_document_type_filename.refresh_from_db()
        self.assertEqual(
            self.test_document_type_filename.filename,
            TEST_DOCUMENT_TYPE_QUICK_LABEL
        )

    def test_document_type_quick_label_edit_with_access(self):
        self.grant_access(
            obj=self.test_document_type,
            permission=permission_document_type_edit
        )
        self._create_test_quick_label()

        response = self._request_quick_label_edit()
        self.assertEqual(response.status_code, 302)

        self.test_document_type_filename.refresh_from_db()
        self.assertEqual(
            self.test_document_type_filename.filename,
            TEST_DOCUMENT_TYPE_QUICK_LABEL_EDITED
        )

    def test_document_type_quick_label_list_no_access(self):
        self._create_test_quick_label()

        response = self._request_quick_label_list_view()
        self.assertEqual(response.status_code, 404)

    def test_document_type_quick_label_list_with_access(self):
        self._create_test_quick_label()

        self.grant_access(
            obj=self.test_document_type,
            permission=permission_document_type_view
        )

        response = self._request_quick_label_list_view()
        self.assertContains(
            response, text=self.test_document_type_filename, status_code=200
        )


class DocumentsQuickLabelViewsTestCase(DocumentTypeQuickLabelTestMixin, GenericDocumentViewTestCase):
    def _request_document_quick_label_edit_view(self, extra_data=None):
        data = {
            'document_type_available_filenames': self.test_document_type_filename.pk,
            'label': ''
            # View needs at least an empty label for quick
            # label to work. Cause is unknown.
        }
        data.update(extra_data or {})

        return self.post(
            viewname='documents:document_edit', kwargs={
                'pk': self.test_document.pk
            }, data=data
        )

    def test_document_quick_label_no_permission(self):
        self._create_test_quick_label()

        response = self._request_document_quick_label_edit_view()
        self.assertEqual(response.status_code, 404)

    def test_document_quick_label_with_access(self):
        self._create_test_quick_label()
        self.grant_access(
            obj=self.test_document,
            permission=permission_document_properties_edit
        )

        response = self._request_document_quick_label_edit_view()
        self.assertEqual(response.status_code, 302)

        self.test_document.refresh_from_db()
        self.assertEqual(
            self.test_document.label, self.test_document_type_filename.filename
        )

    def test_document_quick_label_preserve_extension_with_access(self):
        self._create_test_quick_label()
        self.grant_access(
            permission=permission_document_properties_edit, obj=self.test_document
        )
        filename, extension = os.path.splitext(self.test_document.label)

        response = self._request_document_quick_label_edit_view(
            extra_data={'preserve_extension': True}
        )
        self.assertEqual(response.status_code, 302)

        self.test_document.refresh_from_db()
        self.assertEqual(
            self.test_document.label, '{}{}'.format(
                self.test_document_type_filename.filename, extension
            )
        )

    def test_document_quick_label_no_preserve_extension_with_access(self):
        self._create_test_quick_label()
        self.grant_access(
            obj=self.test_document,
            permission=permission_document_properties_edit
        )
        filename, extension = os.path.splitext(self.test_document.label)

        response = self._request_document_quick_label_edit_view(
            extra_data={'preserve_extension': False}
        )
        self.assertEqual(response.status_code, 302)

        self.test_document.refresh_from_db()
        self.assertEqual(
            self.test_document.label, self.test_document_type_filename.filename
        )
